import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.io.*;
import java.net.*;
import java.util.HashMap;

public class Board extends World
{
    public static final int LEVELS = 100;

    private Paddle player;
    private Paddle support;
    private Ball ball;
    private Counter score = new Counter(0);
    private Counter lives = new Counter(3);
    private Counter level = new Counter(1);
    
    private Color[] colors = { Color.white, Color.magenta, Color.pink, Color.red, Color.orange, Color.yellow, Color.blue, Color.green, Color.cyan, new Color(150, 90, 0), Color.lightGray };
    private char[] literals = { 'w', 'm', 'p', 'r', 'o', 'y', 'b', 'g', 'c', 'x', 'l' };
    private HashMap<String, Integer> passwords = new HashMap<String, Integer>();
    private HashMap<Integer, String> levels = new HashMap<Integer, String>();
    
    private String fallback = getLevel(0);
    private boolean editorMode, paused, keyPressed, keyDown;
    private int difficulty;
    private int undestroyableBricks = 0;

    public Board() {    
        super(600, 520, 1); 
        paused = true;
        setPaintOrder(Button.class, Cushion.class, Ball.class, Paddle.class, Item.class, Brick.class);
        initializePasswords();
        addObject(new Menu(), getWidth()/2, getHeight()/2);
    }
    
    public void start(boolean pMouseControl, boolean pMultiplayer, int pDifficulty) {
        editorMode = false;
        paused = false;
        removeObjects(getObjects(null));
        addGameElements(pMouseControl, pMultiplayer, pDifficulty);
        buildLevel(getLevel(level.getCount()));
    }
    
    public void jumpToLevel(int pLevel, boolean pMouseControl, boolean pMultiplayer, int pDifficulty) {
        if(pLevel > 0 && pLevel <= LEVELS) {
            level = new Counter(pLevel);
            start(pMouseControl, pMultiplayer, pDifficulty);
        }
    }
    
    public void addGameElements(boolean pMouseControl, boolean pMultiplayer, int pDifficulty) {
        difficulty = pDifficulty/3;
        player = new Paddle(pMouseControl, true, pDifficulty);
        support = new Paddle(!pMouseControl, false, pDifficulty);
        addObject(player, (getWidth()-140)/2, getHeight()-40);
        if(pMultiplayer) {
            addObject(support, (getWidth()-140)/2, getHeight()-20);
        }  
        if(!editorMode) {
            addObject(score, 535, 65);
            addObject(lives, 585, 180);
            addObject(level, 585, 300);
        }
        else {
            addObject(new Button("Back", Color.red), 530, 480);
        }
    }
    
    private void buildLevel(String pLevel) {
        int i=0;
        for(int y=10; y < getHeight()-100; y+=21) {
            for(int x=23; x < getWidth()-140; x+=46) {
                if(i < pLevel.length()) {
                    Actor brick = getBrick(pLevel.charAt(i));
                    if(brick != null) {
                        addObject(brick, x, y);
                    }
                }
                i++;
            }
        }
    }
    
    public int getLevel(String pwd) {
         if(passwordExists(pwd)) {
             return passwords.get(pwd);
         }
         return -1;
    }
    
    public boolean passwordExists(String pwd) {
        return passwords.containsKey(pwd);    
    }
    
    private String getLevel(int pLevel) {
        String file = "level/lvl" + pLevel + ".txt";
        String lv = "";
        try {
            URL path = getClass().getClassLoader().getResource(file);
            BufferedReader input = new BufferedReader(new InputStreamReader(path.openStream()));           
            do {
                input.mark(10);
                String line = input.readLine();
                if(line.length() > 10) {
                    line = line.substring(0, 10);
                }
                else {
                    while(line.length() < 10) {
                        line += " ";
                    }
                }
                lv += line;
                input.reset();
            }
            while(input.readLine() != null);
        }
        catch(Exception IOException) { }
        return lv;
    }
    
    private Actor getBrick(char pLiteral) {
        for(int i=0; i < literals.length-2; i++) {
            if(pLiteral == literals[i]) {
                return new Brick(colors[i], pLiteral);
            }
        }     
        if(pLiteral == 'l') {
            return new Brick(Color.lightGray, 3, true, pLiteral);
        }
        else if(pLiteral == 'x') {
            undestroyableBricks++;
            return new Brick(new Color(150, 90, 0), 0, false, pLiteral);           
        }
        else if(pLiteral == 't') {
            return new GridCell();
        }
        else {
            return null;
        }
    }
    
    public void initializePasswords() {
        try {
            URL path = getClass().getClassLoader().getResource("pwds.txt");
            BufferedReader input = new BufferedReader(new InputStreamReader(path.openStream()));
            int lvl = 0;
            do {
                input.mark(5);
                String pwd = input.readLine();
                lvl++;
                passwords.put(pwd, lvl);
                levels.put(lvl, pwd);
                input.reset();
            } while(input.readLine() != null);
        } catch(IOException e) {}
    }
    
    public void setBall(Ball pBall) {
        ball = pBall;
    }
    
    public Ball getBall() {
        return ball;
    }
    
    public void act() {         
        if(!paused) {
            if(getObjects(Brick.class).size() == undestroyableBricks && getObjects(ScoreBoard.class).isEmpty()) {
                levelBeaten();
            }
            if(Greenfoot.isKeyDown("m") && !keyDown) {
                Ball.muteSound();
                Projectile.muteSound();
                keyDown = true;
            }
            if(!Greenfoot.isKeyDown("m") && keyDown) {
                keyDown = false;
            }
        }
        pause();
    }
    
    public void ballIsOut(Ball b) {
        if(b != null) {
            removeObject(b);
        }
        else {
            removeObjects(getObjects(Ball.class));
        }
        if(getObjects(Ball.class).size() <= 0) {
            lives.change(-1);
            if(lives.getCount() > 0 || editorMode) {
                player.newBall();
                support.normalize();
            }
            else {
                askContinue();
            }
        }
    }
    
    public void score(int pScore) {
        score.change(pScore*difficulty);
    }
    
    public void life(int pLife) {
        lives.change(pLife);
    }
    
    private void askContinue() {
        addObject(new Cushion(), (getWidth()-140)/2, getHeight()/2);
        addObject(new Button("Y", Color.white), (getWidth()-140)/2-100, getHeight()/2);
        addObject(new Button("N", Color.white), (getWidth()-140)/2+100, getHeight()/2);
        addObject(new Button("Password: " + levels.get(level.getCount()), Color.white, 50), (getWidth()-140)/2, getHeight()/2+150);
    }
    
    public void restartLevel() {
        removeObjects(getObjects(Cushion.class));
        removeObjects(getObjects(Button.class));
        score.change(-score.getCount());
        lives.change(3);
        level.change(-1);
        levelBeaten();
    }
    
    public void gameOver() {
        setBackground("images/start.png");
        removeObjects(getObjects(null));
        if(UserInfo.isStorageAvailable()) {
            if(UserInfo.getMyInfo() != null) {
                UserInfo player = UserInfo.getMyInfo();
                if(player.getScore() < score.getCount()) {
                    player.setScore(score.getCount());
                    player.store();
                }
            }
        }        
        addObject(new ScoreBoard(getWidth(), getHeight()), getWidth()/2, 230);
    }
    
    public void pause() {
        if(Greenfoot.isKeyDown("enter") && !keyPressed) {
            keyPressed = true;
            paused = !paused;
            Ball.pause();
            Paddle.pause();
            Brick.pause();
            Item.pause();
        }
        else if(!Greenfoot.isKeyDown("enter") && keyPressed) {
            keyPressed = false;            
        }
    }
    
    public void levelEditor() {
        editorMode = true;
        setBackground("images/editor.png");
        removeObjects(getObjects(null));
        int i=0;
        for(int y=15; y <= 95; y+=20) {
            for(int x=485; x <= 575; x+=45) {
                if(i < colors.length) {
                    if(i < colors.length-2)  {  addObject(new EditorBrick(colors[i], 1, true, false, literals[i]), x, y);   }
                    if(i == colors.length-2) {  addObject(new EditorBrick(colors[i], 0, false, false, literals[i]), x, y);  }
                    if(i == colors.length-1) {  addObject(new EditorBrick(colors[i], 3, true, false, literals[i]), x, y);   }
                }  
                i++;
            }
        }
        addObject(new Button("Load", Color.red), 530, 350);
        addObject(new Button("Save", Color.red), 530, 415);
        addObject(new Button("Play!", Color.red), 530, 480);
        buildLevel(fallback);
    }
    
    public String translateLevel() {
        StringBuilder lv = new StringBuilder("");
        if(!getObjects(Brick.class).isEmpty()) {
            for(int y=10; y < getHeight(); y+=21) {
                for(int x=23; x < getWidth()-140; x+=46) {
                    if(getObjectsAt(x, y, Brick.class).isEmpty()) {  
                        if(getObjectsAt(x, y, GridCell.class).isEmpty()) {  lv.append(" ");  }
                        else {  lv.append('t');  }
                    }
                    else {  
                        for(Object b : getObjectsAt(x, y, Brick.class)) {  lv.append(((Brick) b).getLiteral());  }  
                    }
                }
            }
        }
        return lv.toString();
    }
    
    public void saveCustomLevel(String pLevel) {
        if(UserInfo.isStorageAvailable()) {
            if(UserInfo.getMyInfo() != null) {
                UserInfo player = UserInfo.getMyInfo();
                if(!pLevel.equals("")) {
                    for(int i=0; i < UserInfo.NUM_STRINGS; i++) {
                        player.setString(i, pLevel.substring(0+i*UserInfo.STRING_LENGTH_LIMIT, UserInfo.STRING_LENGTH_LIMIT*(i+1)));
                    }
                    player.store();
                }
            }
        }
    }
    
    public void loadCustomLevel() {
        for(Object b : getObjects(Brick.class)) {
            ((Brick) b).remove();
        }
        levelEditor();
        StringBuilder lv = new StringBuilder("");
        if(UserInfo.isStorageAvailable()) {
            if(UserInfo.getMyInfo() != null) {
                UserInfo player = UserInfo.getMyInfo();
                for(int i=0; i < UserInfo.NUM_STRINGS; i++) {
                    lv.append(player.getString(i));
                }
            }
        }
        buildLevel(lv.toString());
    }
    
    public void testLevel() {
        paused = false;
        fallback = translateLevel();
        removeObjects(getObjects(null));
        buildLevel(fallback);
        addGameElements(false, false, 6);
    }
    
    private void levelBeaten() {
        for(Object b : getObjects(Brick.class)) {
            ((Brick) b).remove();
        }
        removeObjects(getObjects(Ball.class));
        removeObjects(getObjects(Item.class));
        removeObjects(getObjects(Net.class));
        undestroyableBricks = 0;
        if(!editorMode) {
            player.setLocation((getWidth()-140)/2, getHeight()-40);
            support.setLocation((getWidth()-140)/2, getHeight()-20);
            player.newBall();
            level.change(1);
            if(level.getCount() <= LEVELS) {  buildLevel(getLevel(level.getCount()));  }
            else {  gameOver();  }
        }
        else {  levelEditor();   }
    }
}